/**
 * Copyright (c) 2009, Martin Lanter
 * All rights reserved.
 */


package ch.lantersoft.layeredPaneExperiment;

import java.awt.AlphaComposite;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.GradientPaint;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.util.Random;

import javax.swing.JButton;
import javax.swing.JLayeredPane;
import javax.swing.Timer;

import org.jdesktop.animation.timing.Animator;
import org.jdesktop.animation.timing.TimingTargetAdapter;

public class LpExperiment extends JLayeredPane implements ComponentListener, MouseWheelListener {

	private static final long serialVersionUID = 8731635437463581316L;

	private Animator animator;
	private Timer timer; // Timer that adds panels
	private Config config;
	private int counter;

	public LpExperiment(final Config config) {
		this.config = config;

		start();

		addComponentListener(this);
		addMouseWheelListener(this);
	}

	/**
	 * create a new, colored panel and add it to the content
	 */
	public void addNewPanel() {
		ColorPanel cp = new ColorPanel(createColor(),this);
		cp.setText("Panel. "+counter);
		add(cp, new Integer(counter++));
		sort();
	}

	/**
	 * crate a new JButton and add it to the content
	 */
	public void addNewButton() {
		JButton button = new JButton("<html>Button<br><center>"+counter+"</center></html>");
		button.setPreferredSize(new Dimension(80,120));
		button.addMouseListener(new DefaultMousePolicy(this,button));
		add(button,new Integer(counter++));
		sort();
	}

	/**
	 * adjust all children components
	 */
	private synchronized void sort() {
		if (animator!=null) {
			animator.stop();
		}
		Component[] cmps = getComponents();
		if (cmps.length==0) return;
		int n = cmps.length;
		int width = getWidth()-20-cmps[n-1].getPreferredSize().width;
		int height = getHeight()-20-cmps[n-1].getPreferredSize().height;
		Rectangle[] target = new Rectangle[n];
		if (n==1) {
			target[0] = new Rectangle(10,10,cmps[0].getPreferredSize().width,cmps[0].getPreferredSize().height); // out of loop because of /n
		} else {
			for (int i=0;i<n;i++) {
				int index = getLayer(cmps[i]);
				target[i] = new Rectangle(
						10+(index)*width/(n-1),
						10+(index)*height/(n-1),
						cmps[i].getPreferredSize().width,
						cmps[i].getPreferredSize().height);
			}
		}
		move(cmps, target);
	}

	/**
	 * move the given Component one position up
	 */
	public synchronized void up(Component c) {
		if (animator!=null) {
			animator.stop();
		}
		int n = getComponentCount();
		Component[] cmps = getComponents();
		Component c2 = null;
		int z1 = getLayer(c);
		int dz = Integer.MAX_VALUE;
		for (int i=0;i<n;i++) {
			if (0 < getLayer(cmps[i])-z1 && getLayer(cmps[i])-z1 < dz) {
				dz = getLayer(cmps[i])-z1;
				c2 = cmps[i];
			}
		}
		if (c2!=null) {
			swap(c,c2);
		} else { // up all the elements and put c down
			firstToBack(c);
		}
	}

	/**
	 * move the given Component one position down
	 */
	public synchronized void down(Component c) {
		if (animator!=null) {
			animator.stop();
		}
		int n = getComponentCount();
		Component[] cmps = getComponents();
		Component c2 = null;
		int z1 = getLayer(c);
		int dz = Integer.MIN_VALUE;
		for (int i=0;i<n;i++) {
			if (0 > getLayer(cmps[i])-z1 && getLayer(cmps[i])-z1 > dz) {
				dz = getLayer(cmps[i])-z1;
				c2 = cmps[i];
			}
		}
		if (c2!=null) {
			swap(c,c2);
		} else { // up all the elements and put c down
			lastToFront(c);
		}
	}

	/**
	 * move all components one position up and the topmost to the bottom
	 */
	private synchronized void upAll() {
		Component[] cmps = getComponents();
		int n = cmps.length;
		int max = getLayer(cmps[0]);
		Component maxc = cmps[0];
		for (int i=1;i<n;i++) {
			if (getLayer(cmps[i])>max) {
				maxc = cmps[i];
				max = getLayer(cmps[i]);
			}
		}
		firstToBack(maxc);
	}

	/**
	 * move all components one position down and the lowest one to the top
	 */
	private synchronized void downAll() {
		Component[] cmps = getComponents();
		int n = cmps.length;
		int min = getLayer(cmps[0]);
		Component minc = cmps[0];
		for (int i=0;i<n;i++) {
			if (getLayer(cmps[i]) < min) {
				minc = cmps[i];
				min = getLayer(cmps[i]);
			}
		}
		lastToFront(minc);
	}

	/**
	 * move the first component to the back
	 */
	private synchronized void firstToBack(Component first) {
		Component[] cmps = getComponents();
		int n = cmps.length;
		int min = Integer.MAX_VALUE;
		for (int i=0;i<n;i++) {
			int z = getLayer(cmps[i]);
			if (z<min) {
				min = z;
			}
			setLayer(cmps[i],z+1);
		}
		setLayer(first,min);
		sort();
	}

	/**
	 * move the last component into the front
	 */
	private synchronized void lastToFront(Component last) {
		Component[] cmps = getComponents();
		int n = cmps.length;
		int max = Integer.MIN_VALUE;
		for (int i=0;i<n;i++) {
			int z = getLayer(cmps[i]);
			if (z > max) {
				max = z;
			}
			setLayer(cmps[i],z-1);
		}
		setLayer(last,max);
		sort();
	}

	/**
	 * move the given components to the given rectangles
	 */
	private synchronized void move(final Component[] cmps, final Rectangle[] target) {
		move(cmps,target,config.time);
	}

	/**
	 * move the given components to the given rectangles within the given time
	 */
	private synchronized void move(final Component[] cmps, final Rectangle[] target, int time) {
		if (cmps==null || target==null || cmps.length!=target.length) return;
		if (animator!=null) {
			animator.stop();
		}
		final int n = cmps.length;
		final Rectangle[] start = new Rectangle[n];
		for (int i=0;i<n;i++) {
			start[i] = cmps[i].getBounds();
		}
		animator = new Animator(time);
		animator.setAcceleration(.4f);
		animator.setDeceleration(.2f);
		animator.addTarget(new TimingTargetAdapter() {
			public void timingEvent(float fraction) {
				for (int i=0;i<n;i++) {
					float sign = Math.signum(target[i].x-start[i].x);
					int x = (int)(start[i].x+(target[i].x-start[i].x)*fraction);
					int y = (int)(start[i].y+(target[i].y-start[i].y)*fraction*fraction*fraction*fraction); // f^4 makes the bow
					int w = (int)((start[i].width+(target[i].width-start[i].width)*fraction) * (config.shrink?(1+sign*config.shrinkfactor*4*(fraction<.5?fraction:1-fraction)):1f));
					int h = (int)((start[i].height+(target[i].height-start[i].height)*fraction) *  (config.shrink?(1+sign*config.shrinkfactor*4*(fraction<.5?fraction:1-fraction)):1f));
					cmps[i].setBounds(x,y,w,h);
					cmps[i].validate();
				}
			}
			public void end() {
				animator = null; // to avoid, animator.stop calls the end one more time
				for (int i=0;i<n;i++) {
					cmps[i].setBounds(target[i]);
				}
				repaint();
			}
		} );
		animator.start();
	}

	/**
	 * switch the two given components' places
	 */
	private synchronized void swap(Component c1, Component c2) {
		int temp = getLayer(c1);
		move(new Component[] {c1, c2}, new Rectangle[] {c2.getBounds(), c1.getBounds()});
		setLayer(c1, getLayer(c2));
		setLayer(c2, temp);
	}

	/**
	 * start the demonstration
	 */
	public synchronized void start() {
		if (timer!=null) timer.stop();

		final boolean shrink = config.shrink;
		config.shrink = false;
		counter = 0;
		timer = new Timer(config.time,null);
		timer.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent event) {
				if (counter<config.panes) {
					addNewPanel();
				} else if (config.panes<=counter && counter<config.panes+config.buttons) {
					addNewButton();
				} else if (counter>=config.panes+config.buttons) {
					timer.stop();
					config.shrink = shrink;
				}
				if (timer.getDelay()!=config.time) {
					timer.setDelay(config.time);
				}
			}
		});
		timer.start();
	}

	/**
	 * clear the content and remove all children
	 */
	public void clear() {
		Animator animator = new Animator(config.time);
		animator.setAcceleration(.3f);
		animator.setDeceleration(.3f);
		animator.addTarget(new TimingTargetAdapter() {
			public void timingEvent(float fraction) {
				config.alpha = 1-fraction;
				repaint();
			}
		    public void begin() { }
		    public void end() {
		    	config.alpha = 1.0f;
				removeAll();
				counter = 0;
		    }
		});
		animator.start();
	}

	/**
	 * randomly create a new color
	 */
	private Color createColor() {
		Random rand = new Random();
		int col = rand.nextInt();
		Color ret = new Color(col%(256*256*256));
		return ret;
	}

	@Override
	public void paintComponent(Graphics g) {
		Graphics2D g2 = (Graphics2D)g;
		GradientPaint gradient = new GradientPaint(0,getHeight(),config.col1
												,getWidth(),getHeight()/2,config.col2);
		g2.setPaint(gradient);
		g2.fill(g.getClipBounds());
		super.paintComponent(g);
	}

	@Override
	public void paintChildren(Graphics g) {
		Graphics2D g2 = (Graphics2D) g.create();
		if (config.alpha<1) {
			g2.setComposite(AlphaComposite.SrcOver.derive(config.alpha)); }
		super.paintChildren(g2);
	}

	@Override
	public void mouseWheelMoved(final MouseWheelEvent event) {
		if (event.getWheelRotation()>0) {
			upAll();
		} else if (event.getWheelRotation()<0) {
			downAll();
		}
	}

	@Override
    public void componentResized(ComponentEvent e) {
    	sort();
    }

	@Override
    public void componentShown(ComponentEvent e) {
    	sort();
    }

	@Override public void componentHidden(ComponentEvent e) { }
	@Override public void componentMoved(ComponentEvent e) { }

}
